// *************************************************************************************************************
// * InfoElektronika Project: Programmable Autonomous Vehicle - PAV                                            *
// * Author: Vladimir Savic                                                                                    *
// * email: vladimirsvc@yahoo.com                                                                              *
// * --------------------------------------------------------------------------------------------------------- *
// * Finite State Machine Definition                                                                           *
// * -------------------------                                                                                 *
// * | Visible States of PAV |                                                                                 *
// * -------------------------                                                                                 *
// * | Stop                  |                                                                                 *
// * | Forward               |                                                                                 *
// * | Backward              |                                                                                 *
// * | RotateLeft            |                                                                                 *
// * | RotateRight           |                                                                                 *
// * -------------------------                                                                                 *
// * Other states are used for internal control software purpose                                               *
// * Sequence of transitions between states:                                                                   *
// * --------------------------------------------------------------------------------------------------------- *
// * Stop; Forward; Stop; Backward; Stop; RotateLeft; Forward; Stop; Backward; Stop; RotateRight; Forward; ... *
// * --------------------------------------------------------------------------------------------------------- *
// * Next state is determined by current sequence, sensor interrupt request or timer interrupt request.        *
// * --------------------------------------------------------------------------------------------------------- *
// * System Development Platform: Easy 8051 v6 for Atmel AT89S8253 MCU                                         *
// * For technical details consult Atmel AT89S8253 MCU specification                                           *
// * Programming Language: mikroC                                                                              *
// *************************************************************************************************************

// *********************************
// * Gloabal variables declaration *
// *********************************

struct{
    ForwardFullDutyCycle  : 1     ; // forward dc motor drive direction state  - maximum speed
    ForwardPWMDutyCycle   : 1     ; // forward dc motor drive direction state  - speed regulation by PWM signal
    BackwardFullDutyCycle : 1     ; // backward dc motor drive direction state - maximum speed
    BackwardPWMDutyCycle  : 1     ; // backward dc motor drive direction state - speed regulation by PWM signal
    TurnRight             : 1     ; // indicate front wheels rotation direction (initial state to the right)
    TurnCounter           : 8     ; // turn direction counter
} STATE_REG;

struct{
    unsigned short ControlSignals ; // stepper motor control signals
    unsigned short AngleCounter   ; // stepper motor angle counter - counting is base on MCU timer0
    unsigned short AnglePreset    ; // stepper motor predefined value (desired value)
} STEPPER_MOTOR;

struct{
    unsigned short Counter        ; // 100% duty cycle counter register, incremental frequency depends on Rohm574 interrupt frequency
                                    // until the counter reaches forward/backward preset value drive dc motor runs on maximum speed
    unsigned short ForwardPreset  ; // 100% duty cycle forward predefined value (desired value)
    unsigned short BackwardPreset ; // 100% duty cycle backward predefined value (desired value)
} DRIVE_DC_MOTOR_FULL_DUTY_CYCLE;

sbit START_TIMER   at TR0_bit     ; // start timer0 control bit, b1 - ON, b0 - OFF

//           .        Control Signals
//           .                           -----------    
//           |- P2.0 ------->---------- |1B       1C|-- stepper motor control signal #1 (Q1)
//           |- P2.1 ------->---------- |2B   U   2C|-- stepper motor control signal #2 (Q2)
//    MCU    |- P2.2 ------->---------- |3B   L   3C|-- stepper motor control signal #3 (Q3)
//           |- P2.3 ------->---------- |4B   N   4C|-- stepper motor control signal #4 (Q4)
//           |- P2.4 ------->---------- |5B   2   5C|-- LED forward direction indicator
//           |- P2.5 ------->---------- |6B   8   6C|-- PWM forward direction (H-Bridge part#1)
//           |- P2.6 ------->---------- |7B   0   7C|-- LED backward direction indicator
// ....------|- P2.7 ------->---------- |8B   3   8C|-- PWM backward direction (H-Bridge part#2)
//                                      |GND     COM|
//                                       -----------
sbit Q1_H_BRIDGE_SIGNAL  at P2_5_bit    ; // PWM signal for Q1 transistor, forward state  - signal for H-Bridge (drive DC motor controller)
sbit Q2_H_BRIDGE_SIGNAL  at P2_7_bit    ; // PWM signal for Q2 transistor, backward state - signal for H-Bridge (drive DC motor controller)
sbit SIGNAL_F_LED        at P2_4_bit    ; // Indicator LED for forward direction
sbit SIGNAL_B_LED        at P2_6_bit    ; // Indicator LED for backward direction

// ***************************
// * Main entry point method *
// ***************************
void main() {
     P0     = 0xFF; // set port 0
     P1     = 0xFF; // set port 1
     P2     = 0x00; // set port 2, ULN2803 device driver is conected to port 2
     P3     = 0xFF; // set port 3

     STATE_REG.ForwardFullDutyCycle  = b1  ; // ON
     STATE_REG.ForwardPWMDutyCycle   = b0  ; // OFF
     STATE_REG.BackwardFullDutyCycle = b0  ; // OFF
     STATE_REG.BackwardPWMDutyCycle  = b0  ; // OFF
     STATE_REG.TurnRight             = b1  ; // ON
     STATE_REG.TurnCounter           = 0x00; // reset turn counter

     DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter        = 0x00; // reset full duty cycle counter
     DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.ForwardPreset  = 0x1E; // preset forward number
     DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.BackwardPreset = 0x06; // preset backward number
     
     STEPPER_MOTOR.ControlSignals = 0x00;  // set current state of stepper motor control signals
     STEPPER_MOTOR.AngleCounter   = 0x00;  // reset stepper motor angle counter to zero
     STEPPER_MOTOR.AnglePreset    = 0x30;  // preset stepper motor angle

     // Timer 0 Mode Select Bit
     // M10 M00 Operating mode
     //  0   0  Mode 0: 8-bit timer/counter (TH0) with 5-bit prescaler (TL0).
     //  0   1  Mode 1: 16-bit timer/counter.
     //  1   0  Mode 2: 8-bit auto-reload timer/counter (TL0). Reloaded from TH0 at overflow.
     //  1   1  Mode 3: TL0 is an 8-bit timer/counter.
     M00_bit = 1;
     M10_bit = 0;

                   // interrupt Enable (IE) Register:
     EA_bit  = 1;  // if EA = 0, no interrupt is acknowledged. If EA = 1, each interrupt source is individually
                   // enabled or disabled by setting or clearing its enable bit
     EX0_bit = 1;  // external interrupt 0 enable bit - EXT#0
     EX1_bit = 1;  // external interrupt 1 enable bit - EXT#1
     ET0_bit = 1;  // timer 0 interrupt enable bit
                   // interrupt Priority (IP) Register
     PX0_bit = 1;  // external Interrupt 0 Priority Low
     PX1_bit = 1;  // external Interrupt 1 Priority Low
     PT0_bit = 0;  // timer 0 Interrupt Priority Low
                   // interrupt Type Control Bits
     IT0_bit = 1;  // set to select falling edge active (edge triggered) for external interrupt 0 - EXT#0
     IT1_bit = 1;  // set to select falling edge active (edge triggered) for external interrupt 1 - EXT#1
}

// **************************************************
// * Start timer0 for puls-width signal generation  *
// **************************************************
void StartTimerForPWMGeneration(){
    TH0 = 255;
    TL0 = 0;
    START_TIMER = b1;
}

// *************************************************************************
// * This method sets wheels direction.                                    *
// * all wheel direction indicators are set to OFF at initialization time  *
// *************************************************************************
void SetWheelsDirection(){
     STATE_REG.TurnCounter++;
     if(STATE_REG.TurnCounter == 0x02){
         STATE_REG.TurnCounter = 0x00;
         STATE_REG.TurnRight = !STATE_REG.TurnRight;
     }
}

// **********************************************************************
// * Interrupt vector service rutine for external intterupt EXT#0       *
// * Triggered by following sensors: IR sensor and front micro switches *
// **********************************************************************
void Int0InterruptHandler() org IVT_ADDR_EX0 {
     // transition: Forward state to Backward state
     if(STATE_REG.ForwardFullDutyCycle || STATE_REG.ForwardPWMDutyCycle){
         START_TIMER = b0;
         P2 = 0x00;
         Q2_H_BRIDGE_SIGNAL = b1;
         DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter = 0x00;
         STATE_REG.ForwardFullDutyCycle  = b0;
         STATE_REG.ForwardPWMDutyCycle   = b0;
         STATE_REG.BackwardFullDutyCycle = b1;
     }
}

// ***********************************************************************
// * Interrupt vector service rutine for external intterupt EXT#1        *
// * Triggered by following sensors: rear and Rohm 574 photo interrupter *
// ***********************************************************************
void Int1InterruptHandler() org IVT_ADDR_EX1 {
    // 100% duty cycle controller
    // it is used only for two states: Forward and Backward direction
    // DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter register increment frequency depends on Rohm574 sensor interrupt frequency
    if(STATE_REG.ForwardFullDutyCycle){
        Q1_H_BRIDGE_SIGNAL = b1;                                                   // force PWM Forward signal to 100% duty cycle
        SIGNAL_F_LED = b1;                                                         // forward direction LED indicator
        DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter++;                                  // increment 100% duty cycle counter register
        if(DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter == DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.ForwardPreset){
            DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter = 0x00;                         // at this point 100% duty cycle is no longer available
            STATE_REG.ForwardFullDutyCycle = b0;                                   // for drive DC motor. duty cycle is reduced to slow down drive DC motor
            STATE_REG.ForwardPWMDutyCycle  = b1;                                   // complete control is done by interrupt timer 0 rutine
            StartTimerForPWMGeneration();
        }
    }
    else if(STATE_REG.BackwardFullDutyCycle){
        Q2_H_BRIDGE_SIGNAL = b1;                                                   // force PWM Backward signal to 100% duty cycle
        SIGNAL_B_LED = b1;                                                         // backward direction LED indicator
        DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter++;                                  // increment 100% duty cycle counter register
        if(DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter == DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.BackwardPreset){
            DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter = 0x00;                         // at this point 100% duty cycle is no longer available
            STATE_REG.BackwardFullDutyCycle = b0;                                  // for drive DC motor. duty cycle is reduced to slow down drive DC motor
            STATE_REG.BackwardPWMDutyCycle  = b1;                                  // complete control is done by interrupt timer 0 rutine
            StartTimerForPWMGeneration();
        }
    }
}

// *********************************************************************
// * Pulse-width modulation signal generation                          *
// * width is detemined by timer0 value                                *
// * when timer value has been reached, Q1_H_BRIDGE_SIGNAL is inverted *
// *********************************************************************
void ForwardPWMSignalGenerator(){
    Q1_H_BRIDGE_SIGNAL = !Q1_H_BRIDGE_SIGNAL;
    TH0 = 115;
    TL0 = 0;
    START_TIMER = b1;
}

// *********************************************************************
// * Pulse-width modulation signal generation                          *
// * width is detemined by timer0 value                                *
// * when timer value has been reached, Q2_H_BRIDGE_SIGNAL is inverted *
// *********************************************************************
void BackwardPWMSignalGenerator(){
    Q2_H_BRIDGE_SIGNAL = !Q2_H_BRIDGE_SIGNAL;
    DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter++;
    if(DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter == 0x07){
        DRIVE_DC_MOTOR_FULL_DUTY_CYCLE.Counter = 0x00;
        P2 = 0x00;
        STATE_REG.BackwardPWMDutyCycle = b0;
        SetWheelsDirection();
    }
    TH0 = 50;
    TL0 = 0;
    START_TIMER = b1;
}

// *************************************************************************************
// * Stepper motor control signal generation                                           *
// * 4 x transistors (Q1, Q2, Q3 and Q4 - ULN 2803 IC) are used to drive stepper motor *
// * only one transistor is active per generated signal:                               *
// * Time 1, Timer 2, Time 3, Time 4, Time 5, ...                                      *
// *  Q1      Q2       Q3      Q4      Q1     ...       RIGHT DIRECTION                *
// *  Q4      Q3       Q2      Q1      Q4     ...       LEFT  DIRECTION                *
// * width of the signal is determined by timer0 value                                 *
// * to shift signals in time use bitwise shift c operator                             *
// * STEPPER_MOTOR.ControlSignals register store current stepper motor position        *
// *************************************************************************************
void StepperMotorController(){
  if(STATE_REG.TurnRight){
     if(STEPPER_MOTOR.ControlSignals == 0x00)
         // reset stepper motor control signals to initial state
         STEPPER_MOTOR.ControlSignals = 0x01;
     else if(STEPPER_MOTOR.ControlSignals == 0x08)
         // reset stepper motor control signals to initial state
         STEPPER_MOTOR.ControlSignals = 0x01;
     else
         STEPPER_MOTOR.ControlSignals <<= 1;
  }
  else{
     if(STEPPER_MOTOR.ControlSignals == 0x01)
         // reset stepper motor control signals to initial state
         STEPPER_MOTOR.ControlSignals = 0x08;
     else
         STEPPER_MOTOR.ControlSignals >>= 1;
  }

  // in order to control stepper motor position
  // set new value of control signals to port 2
  P2 = STEPPER_MOTOR.ControlSignals;

  STEPPER_MOTOR.AngleCounter++;                                 // increment current angle of stepper motor
  if(STEPPER_MOTOR.AngleCounter == STEPPER_MOTOR.AnglePreset){  // compare current angle with desired
     START_TIMER = b0;                                          // stop timer0
     P2 = 0x00;                                                 // reset port 2, stepper motor is OFF as well as H-Bridge and LED indicators
     STEPPER_MOTOR.AngleCounter     = 0x00;                     // reset stepper current angle to zero
     STATE_REG.ForwardFullDutyCycle = b1  ;                     // set PAV to forward direction - full duty cycle state
     Q1_H_BRIDGE_SIGNAL = b1;                                   // start generation of full duty cycle signal
  }
  else{
     TH0 = 100;
     TL0 = 0;
     START_TIMER = b1;
  }
}

// **************************************************************************
// * Interrupt vector service rutine for internal timer0 intterupt request  *
// * Triggered by MCU timer0                                                *
// **************************************************************************
void Timer0InterruptHandler() org IVT_ADDR_ET0 {
     START_TIMER = b0;
     if(STATE_REG.ForwardPWMDutyCycle){
        ForwardPWMSignalGenerator();
     }
     else if(STATE_REG.BackwardPWMDutyCycle){
        BackwardPWMSignalGenerator();
    }
    else{
        StepperMotorController();
    }
}
