// *************************************************************************************************************
// * InfoElektronika Project: Programmable Autonomous Vehicle - PAV                                            *
// * Author: Vladimir Savic                                                                                    *
// * email: vladimirsvc@yahoo.com                                                                              *
// * Task: movement defined by black tape                                                                      *
// * --------------------------------------------------------------------------------------------------------- *
// * Finite State Machine Definition                                                                           *
// * -----------------------------                                                                             *
// * | Visible States of PAV     |                                                                             *
// * -----------------------------                                                                             *
// * | Stop                      |                                                                             *
// * | Forward                   |                                                                             *
// * | SearchBlackTrack/Rotation |                                                                             *
// * -----------------------------                                                                             *
// * Other states are used for internal control software purpose                                               *
// * Sequence of transitions between states:                                                                   *
// * --------------------------------------------------------------------------------------------------------- *
// * Example: Stop; SearchBlackTrack; Forward; SearchBlackTrack; Forward; Stop;                                *
// * --------------------------------------------------------------------------------------------------------- *
// * Next state is determined by current sequence, sensor interrupt request or timer interrupt request.        *
// * --------------------------------------------------------------------------------------------------------- *
// * System Development Platform: Easy 8051 v6 for Atmel AT89S8253 MCU                                         *
// * For technical details consult Atmel AT89S8253 MCU specification                                           *
// * Programming Language: mikroC - Advanced C compiler for 8051 Standard                                      *
// *************************************************************************************************************

// Finite State Machine Graph
// ****************************************************************************************************************
// *  -------------                      -----------------------------------                      --------------  *
// *  |           |  Right Micro Switch  |                                 |   Black Track End    |            |  *
// *  |    Stop   |--------------------->| BlackTrackSearch/WheelsRotation |--------------------->|  End/Stop  |  *
// *  |           |                      |                                 |                      |            |  *
// *  -------------                      -----------------------------------                      --------------  *
// *                                                    ^    | IR Sensor/Black Track                              *
// *                                                    |    |   -------------                                    *
// *                            IR Sensor/White Surface |    |-->|           |                                    *
// *                                                    |        |  Forward  |                                    *
// *                                                    |--------|           |                                    *
// *                                                             -------------                                    *
// ****************************************************************************************************************

// *********************************
// * Gloabal variables declaration *
// *********************************

enum { Stop = 0, Forward, SearchBlackTrack };   // Programmable Autonomous Vehicle States
unsigned short currentState ;                   // current state of the vehicle
unsigned short pwmCounter   ;                   // PWM counter

struct{
    TurnDirection : 1          ; // indicate front wheels rotation direction (initial state to the right)
    unsigned short TurnCounter ; // turn direction counter
} STATE_REG;

struct{
    unsigned short ControlSignals ; // stepper motor control signals
    unsigned short AngleCounter   ; // stepper motor angle counter - counting is base on MCU timer0
    unsigned short AnglePreset    ; // stepper motor predefined value (desired value)
    unsigned short AngleCoeff     ; // stepper motor angle coefficient, can be 1 or 2
} STEPPER_MOTOR;

sbit START_TIMER   at TR0_bit     ; // start timer0 control bit, b1 - ON, b0 - OFF

//           .        Control Signals
//           .                           -----------
//           |- P2.0 ------->---------- |1B       1C|-- stepper motor control signal #1 (Q1)
//           |- P2.1 ------->---------- |2B   U   2C|-- stepper motor control signal #2 (Q2)
//    MCU    |- P2.2 ------->---------- |3B   L   3C|-- stepper motor control signal #3 (Q3)
//           |- P2.3 ------->---------- |4B   N   4C|-- stepper motor control signal #4 (Q4)
//           |- P2.4 ------->---------- |5B   2   5C|-- LED forward direction indicator
//           |- P2.5 ------->---------- |6B   8   6C|-- PWM forward direction (H-Bridge part#1)
//           |- P2.6 ------->---------- |7B   0   7C|-- LED backward direction indicator
// ....------|- P2.7 ------->---------- |8B   3   8C|-- PWM backward direction (H-Bridge part#2)
//                                      |GND     COM|
//                                       -----------
sbit Q1_H_BRIDGE_SIGNAL  at P2_5_bit    ; // PWM signal for Q1 transistor, forward state  - signal for H-Bridge (drive DC motor controller)
sbit Q2_H_BRIDGE_SIGNAL  at P2_7_bit    ; // PWM signal for Q2 transistor, backward state - signal for H-Bridge (drive DC motor controller)
sbit SIGNAL_F_LED        at P2_4_bit    ; // Indicator LED for forward  direction
sbit SIGNAL_B_LED        at P2_6_bit    ; // Indicator LED for backward direction
sbit ON_THE_TRACK        at P3_2_bit    ; // Black track sensor; 0 logic level -> no track; 1 logic level -> track

// ***************************
// * Main entry point method *
// ***************************
void main() {
     P0     = 0xFF;                        // set port 0
     P1     = 0xFF;                        // set port 1
     P2     = 0x00;                        // set port 2, ULN2803 device driver is conected to port 2
     P3     = 0xFF;                        // set port 3
     
     pwmCounter   = 0x00 ;                 // reset PWM counter
     currentState = Stop ;                 // S0 state is Stop

     STATE_REG.TurnDirection      = b1  ;  // ON
     STATE_REG.TurnCounter        = 0x00;  // reset turn counter

     STEPPER_MOTOR.ControlSignals = 0x00;  // set current state of stepper motor control signals
     STEPPER_MOTOR.AngleCounter   = 0x00;  // reset stepper motor angle counter to zero
     STEPPER_MOTOR.AnglePreset    = 0x55;  // preset stepper motor angle
     STEPPER_MOTOR.AngleCoeff     = 0x01;  // set angle coefficient to 1

     // Timer 0 Mode Select Bit
     // M10 M00 Operating mode
     //  0   0  Mode 0: 8-bit timer/counter (TH0) with 5-bit prescaler (TL0).
     //  0   1  Mode 1: 16-bit timer/counter.
     //  1   0  Mode 2: 8-bit auto-reload timer/counter (TL0). Reloaded from TH0 at overflow.
     //  1   1  Mode 3: TL0 is an 8-bit timer/counter.
     M00_bit = 1;
     M10_bit = 0;

                   // Interrupt Enable (IE) Register:
     EA_bit  = 1;  // if EA = 0, no interrupt is acknowledged. If EA = 1, each interrupt source is individually
                   // enabled or disabled by setting or clearing its enable bit
     EX0_bit = 1;  // external interrupt 0 enable bit - EXT#0
     EX1_bit = 1;  // external interrupt 1 enable bit - EXT#1
     ET0_bit = 1;  // timer 0 interrupt enable bit
                   // interrupt Priority (IP) Register
     PX0_bit = 1;  // external Interrupt 0 Priority Low
     PX1_bit = 1;  // external Interrupt 1 Priority Low
     PT0_bit = 0;  // timer 0 Interrupt Priority Low
                   // interrupt Type Control Bits
     IT0_bit = 1;  // set to select falling edge active (edge triggered) for external interrupt 0 - EXT#0
     IT1_bit = 1;  // set to select falling edge active (edge triggered) for external interrupt 1 - EXT#1
}

// *************************************************************
// * This method is used to start timer 0 and set timer values *
// *************************************************************
void StartTimer0(unsigned short th0, unsigned short tl0){
    TH0 = th0;
    TL0 = tl0;
    START_TIMER = b1;
}

// **************************************************************
// * This method is used to stop timer 0 and reset timer values *
// **************************************************************
void StopTimer0(){
    TH0 = 0;
    TL0 = 0;
    START_TIMER = b0;
}

// **********************************************************************
// * Interrupt vector service rutine for external intterupt EXT#0       *
// * Triggered by following sensors: IR sensor and front micro switches *
// **********************************************************************
void Int0InterruptHandler() org IVT_ADDR_EX0 {
    // transition function:
    switch(currentState){                     // check current state
        case Forward:
            StopTimer0();                     // stop timer and reset timer values
            Q1_H_BRIDGE_SIGNAL = b0;          // stop drive DC motor
            SIGNAL_F_LED       = b0;          // turn off LED Forward LED indicator
            EX0_bit            = b0;          // disable all interrupts
            currentState = SearchBlackTrack;  // set new state to search black track
            StartTimer0(100, 0);              // start timer and set timer values
        break;
    };
}

// ***********************************************************************
// * Interrupt vector service rutine for external intterupt EXT#1        *
// * Triggered by following sensors: rear and Rohm 574 photo interrupter *
// ***********************************************************************
void Int1InterruptHandler() org IVT_ADDR_EX1 {
    // transition function:
    switch(currentState){                     // check current state
        case Stop:
            currentState = Forward;           // set current state to forward
            Q1_H_BRIDGE_SIGNAL = b1;          // start drive DC motor
            SIGNAL_F_LED       = b1;          // turn on LED Forward LED indicator
            StartTimer0(0, 0);                // start timer and set timer values
        break;
    };
}

// *************************************************************************************
// * Stepper motor control signal generation                                           *
// * 4 x transistors (Q1, Q2, Q3 and Q4 - ULN 2803 IC) are used to drive stepper motor *
// * only one transistor is active per generated signal:                               *
// * Time 1, Timer 2, Time 3, Time 4, Time 5, ...                                      *
// *  Q1      Q2       Q3      Q4      Q1     ...       RIGHT DIRECTION                *
// *  Q4      Q3       Q2      Q1      Q4     ...       LEFT  DIRECTION                *
// * signal width is determined by timer0 value                                        *
// * to shift signals in time use bitwise shift c operator                             *
// * STEPPER_MOTOR.ControlSignals register store current stepper motor position        *
// *************************************************************************************
void StepperMotorController(){
  // stepper motor rotation controller
  if(STATE_REG.TurnDirection){
     if(STEPPER_MOTOR.ControlSignals == 0x00 ||
        STEPPER_MOTOR.ControlSignals == 0x08)
         // reset stepper motor control signals to initial state
         STEPPER_MOTOR.ControlSignals = 0x01;
     else
         // shift signals in time
         STEPPER_MOTOR.ControlSignals <<= 1;
  }
  else{
     if(STEPPER_MOTOR.ControlSignals == 0x01)
         // reset stepper motor control signals to initial state
         STEPPER_MOTOR.ControlSignals = 0x08;
     else
         // shift signals in time
         STEPPER_MOTOR.ControlSignals >>= 1;
  }

  // in order to rotate stepper motor set stepper motor control signals to port 2
  P2 = STEPPER_MOTOR.ControlSignals;

  // after rotation, check state of IR sensor (P3.2 MCU pin)
  // if 0 logic level is presented on MCU P3.2 pin -> no black track    -> continue with wheels rotation
  // if 1 logic level is presented on MCU P3.2 pin -> black track found -> do transition between states -> go to Forward state
  if(!ON_THE_TRACK){
      STEPPER_MOTOR.AngleCounter++;                                                            // increment current angle of stepper motor
      if(STEPPER_MOTOR.AngleCounter == STEPPER_MOTOR.AngleCoeff * STEPPER_MOTOR.AnglePreset){  // compare current angle with desired
         P2 = 0x00;                                                                            // reset port 2, stepper motor is OFF as well as H-Bridge and LED indicators
         STEPPER_MOTOR.AngleCounter = 0x00;                                                    // reset stepper motor angle counter to zero
         if(STEPPER_MOTOR.AngleCoeff == 0x01){
            STEPPER_MOTOR.AngleCoeff = 0x02;                                                   // set angle coeficient to 2, coeficient 2 is full IR sensor black track search
         }                                                                                     //                            coeficient 1 is half IR sensor black track search (initial state - IR sensor in the same direction as the vehicle - turn angle 0 )
         STATE_REG.TurnDirection = !STATE_REG.TurnDirection;                                   // invert turn direction of the wheels
         STATE_REG.TurnCounter++;                                                              // increment turn counter
      }
      if(STATE_REG.TurnCounter < 0x02)
          StartTimer0(100, 0);
  }
  else {
      // black track detected by IR sensor
      // do transition function: go to Forward state and stay there until IR sensor detects white surface
      STATE_REG.TurnCounter = 0x00;      // reset turn counter
      EX0_bit = b1;                      // enable all interrupts
      currentState = Forward;            // set current state to Forward
      Q1_H_BRIDGE_SIGNAL = b1;           // start drive DC motor
      SIGNAL_F_LED       = b1;           // turn off LED Forward LED indicator
      StartTimer0(0, 0);                 // start timer 0 and set timer values
  }
}

// *************************************************************************************
// * This method is used to generate PWM signal for drive DC motor                     *
// *************************************************************************************
void ForwardPWMSignalGenerator(){
    pwmCounter++;                                // increment PWM signal counter
    if(pwmCounter == 0x0B){                      // check desired signal width
      Q1_H_BRIDGE_SIGNAL = !Q1_H_BRIDGE_SIGNAL;  // invert signal to opposite state
    
      if(!Q1_H_BRIDGE_SIGNAL){                   // check voltage level of the PWM signal
        pwmCounter = 0x08;                       // for low level voltage set PWM counter to 0x08
      }                                          // otherwise set PWM counter to 0x00
      else{
        pwmCounter = 0x00;
      }
    }
    StartTimer0(0, 0);                           // start timer and set timer values
}

// **************************************************************************
// * Interrupt vector service rutine for internal timer0 intterupt request  *
// * Triggered by MCU timer0                                                *
// **************************************************************************
void Timer0InterruptHandler() org IVT_ADDR_ET0 {
    StopTimer0();
    switch(currentState){
        case SearchBlackTrack:
            StepperMotorController();
        break;
        case Forward:
            ForwardPWMSignalGenerator();
        break;
    };
}
